"""
atpthings.database.data_entity
-------------------------
"""
from __future__ import annotations

import os
import json
import typing
import pandas as pd
import datetime as dt

from uuid import uuid4, UUID
from pymongo import MongoClient
from bson.json_util import dumps

old_default = json.JSONEncoder.default


def new_default(self, obj):
    if isinstance(obj, UUID):
        return str(obj)
    return old_default(self, obj)


json.JSONEncoder.default = new_default


def data_to_df(data):
    return pd.DataFrame(data)


def data_to_df_ts(data):
    df = pd.DataFrame(data)
    try:
        df.set_index("ts", inplace=True)
        pass
    except Exception:
        pass
    return df


def data_to_dict(data):
    return data


def data_to_str(data):
    return json.loads(json.dumps(data, default=str))


class DataTimeseries:
    def __init__(self, uri: str = None, user: str = None, password: str = None) -> None:
        self.uri = uri
        self.user = user
        self.password = password
        self.database = "entities"
        self.collection = "timeseries"

        if self.uri is None:
            self.uri = os.getenv("MONGODB_ATP_CLOUD_URI")
        if self.user is None:
            self.user = os.getenv("MONGODB_ATP_CLOUD_USER")
        if self.password is None:
            self.password = os.getenv("MONGODB_ATP_CLOUD_PASSWORD")

        self.connect()
        pass

    def connect(self):
        self.db_client = MongoClient(
            self.uri,
            username=self.user,
            password=self.password,
            uuidRepresentation="standard",
        )
        self.data_timeseries_collection = self.db_client[self.database][self.collection]
        return self

    def insert(self, data: dict | list | pd.DataFrame):
        if type(data) == dict:
            data["ts"]
            data["uuid"]
            return self.data_timeseries_collection.insert_one(data)
        elif type(data) == list:
            # TODO: check for "ts" and "uuid"
            return self.data_timeseries_collection.insert_many(data)
        elif type(data) == pd.DataFrame:
            # check for colum names
            df = data.copy()  # check if this is needed
            df.reset_index(inplace=True)
            document = df.to_dict("records")
            return self.data_timeseries_collection.insert_many(document)

    def find(
        self,
        uuid: UUID | list(UUID),
        ts_from: dt.datetime = None,
        ts_to: dt.datetime = None,
        output: pd.DataFrame | dict | str = None,
    ):
        if type(uuid) == UUID:
            list(uuid)

        filter: dict = {"$and": [{"uuid": {"$in": uuid}}]}

        if ts_from:
            filter["$and"].append({"ts": {"$gte": ts_from}})

        if ts_to:
            filter["$and"].append({"ts": {"$lte": ts_to}})

        ret = self.data_timeseries_collection.find(
            filter=filter,
            projection={
                "_id": 0,
            },
        )
        data = list(ret)

        if output == pd.DataFrame:
            return data_to_df_ts(data)
        elif output == dict:
            return data_to_dict(data)
        elif output == str:
            return data_to_str(data)
        else:
            return ret

    def delete(self, uuid: UUID):
        return self.data_timeseries_collection.delete_many({"uuid": uuid})



"""
atpthings.dict
--------------
Utilities for manipulating with dict (dictionary)


"""
from .get import get_keys, get_values, get_keys_as_list, get_values_from_dict_list
from .io import load, save
from .convert import list_to_dict



"""
atpthings.dict.get
------------------
"""


def get_keys(dict_: dict, keys: list) -> dict:
    """Get keys from dictionary.

    Parameters
    ----------
    dict_ : dict
        Dictionary.
    keys : list
        List of keys to be extracted.

    Returns
    -------
    dict
        Dictionary with only specificated keys.

    Examples
    --------

    >>> dict_in = {"one": 1, "two": 2, "three": 3}
    >>> keys_list = ["one", "three"]
    >>> atpthings.util.dictionary.get_keys(dict_in, keys_list)
    >>> {'one': 1, 'three': 3}
    """

    return {key: dict_[key] for key in dict_.keys() & keys}


def get_keys_as_list(dict_: dict) -> list:
    """Get keys from dictionary as list.

    Parameters
    ----------
    dict_ : dict
                Dictionary.

    Returns
    -------
    list
        List of all keys in dictionary (1. level).
    """
    return [key for key in dict_.keys()]


def get_values(dict_: dict, keys: list) -> list:
    """Get values of specific keys from dictionary to list.

    Parameters
    ----------
    dict_ : dict
        Dictionary.
    keys : list
        List of keys wonted to be extracted.

    Returns
    -------
    dict
        List of values with specificated key.

    Examples
    --------

    >>> dict_in = {"one": 1, "two": 2, "three": 3}
    >>> keys_list = ["one", "three"]
    >>> atpthings.util.dictionary.get_values(dict_in, keys_list)
    >>> [1, 3]
    """
    # TODO: support for None keys list -> return all values

    return [dict_[item] for item in keys]


# List of dicts
def get_values_from_dict_list(dict_list: list, key: str, unique: bool = False) -> list:
    """Get key values from list of dictionaries.

    Parameters
    ----------
    dict_list : list
        List of dictionaries
    key : str
        Key to be searched.
    unique : bool, optional (default False)
        Output uniques.

    Returns
    -------
    list
        List of key values.
    """
    # TODO: What to do if key is missing

    list_ = []
    for item in dict_list:
        list_.append(item[key])

    if unique:
        return list(set(list_))

    return list_



"""
atpthings.database.data_entity
-------------------------
"""
from __future__ import annotations

import os
import json
import typing
import pandas as pd

from uuid import uuid4, UUID
from pymongo import MongoClient
from bson.json_util import dumps

old_default = json.JSONEncoder.default


def new_default(self, obj):
    if isinstance(obj, UUID):
        return str(obj)
    return old_default(self, obj)


json.JSONEncoder.default = new_default


def data_to_df(data):
    return pd.DataFrame(data)


def data_to_dict(data):
    return data


def data_to_str(data):
    return json.loads(json.dumps(data, default=str))


class DataEntity:
    def __init__(self, uri: str = None, user: str = None, password: str = None) -> None:
        self.uri = uri
        self.user = user
        self.password = password
        self.database = "entities"
        self.collection = "data"

        if self.uri is None:
            self.uri = os.getenv("MONGODB_ATP_CLOUD_URI")
        if self.user is None:
            self.user = os.getenv("MONGODB_ATP_CLOUD_USER")
        if self.password is None:
            self.password = os.getenv("MONGODB_ATP_CLOUD_PASSWORD")

        self.connect()
        pass

    def connect(self):
        self.db_client = MongoClient(
            self.uri,
            username=self.user,
            password=self.password,
            uuidRepresentation="standard",
        )
        self.data_entity_collection = self.db_client[self.database][self.collection]
        return self

    def new(self, data: dict):
        data["type"]
        data["uuid"] = uuid4()
        return self.data_entity_collection.insert_one(data)

    def find(
        self,
        *args: typing.Any,
        projection: dict = {"_id": False},
        output: pd.DataFrame | dict | str = None,
        **kwargs: typing.Any,
    ):
        ret = self.data_entity_collection.find(*args, projection, **kwargs)
        data = list(ret)

        if output == pd.DataFrame:
            return data_to_df(data)
        elif output == dict:
            return data_to_dict(data)
        elif output == str:
            return data_to_str(data)
        else:
            return ret

    def delete(self, uuid: UUID):
        filter: dict = {"uuid": uuid}
        return self.data_entity_collection.delete_one(filter)



"""
atpthings.dict.convert
----------------------
"""


def list_to_dict(list_of_dict: list, key: str, value: str = None) -> dict:
    """Convert list of dictionaries to dictionary by key.

    Parameters
    ----------
    list_of_dict : list
        List of dictionaries.
    key : str
        Key which value will be used for primary key of dict.
    value : str, optional (default whole dictionary)
        The key in dictionary to be copied as a vlue.

    Returns
    -------
    dict
        Dictionary created from list of dictionaries.
    """
    dict_ = {}
    duplicates_counter = 0
    for item in list_of_dict:
        key_name = item[key]
        if key_name in dict_:
            key_name = f"{key_name}_{duplicates_counter}"
            duplicates_counter += 1
        if value is None:
            dict_[key_name] = item
        else:
            dict_[key_name] = item[value]

    return dict_



"""
atpthings.dict.io
-----------------
"""
from __future__ import annotations
from pathlib import Path
import json
import yaml

# Help: https://stackoverflow.com/questions/1773805/how-can-i-parse-a-yaml-file-in-python


# load
def load(path: str, file_type: str = None) -> dict:
    """Load dictionary from file.

    Parameters
    ----------
    path : str
        Path to file.

    Returns
    -------
    dict
        Dictionary
    """

    if file_type is None:
        file_type = Path(path).suffix.split(".")[-1]

    match file_type.lower():
        case "json":
            return _load_json(path)
        case ("yml" | "yaml"):
            return _load_yml(path)
        case "ini":
            # import configparser
            raise Exception(f"File type ({file_type}) not supported yet.")
        case "xml":
            raise Exception(f"File type ({file_type}) not supported yet.")
        case "pickle":
            raise Exception(f"File type ({file_type}) not supported yet.")
        case "env":
            raise Exception(f"File type ({file_type}) not supported yet.")
        case _:
            raise Exception("File type not supported.")


def save(dict_: dict, path: str, file_type: str = None) -> None:
    """Save dictionary to file.

    Parameters
    ----------
    dict_ : dict
        Dictionary.
    path : str
        Path to file.
    """

    if file_type is None:
        file_type = Path(path).suffix.split(".")[-1]

    match file_type.lower():
        case "json":
            return _save_json(dict_, path)
        case ("yml" | "yaml"):
            return _save_yml(dict_, path)
        case "ini":
            # import configparser
            raise Exception(f"File type ({file_type}) not supported yet.")
        case "xml":
            raise Exception(f"File type ({file_type}) not supported yet.")
        case "pickle":
            raise Exception(f"File type ({file_type}) not supported yet.")
        case "env":
            raise Exception(f"File type ({file_type}) not supported yet.")
        case _:
            raise Exception(f"File type ({file_type}) not supported.")


# json
def _load_json(path: str) -> dict:
    with open(path) as file_:
        ret = dict(json.load(file_))

    return ret


def _save_json(dict_: dict, path: str) -> None:

    with open(path, "w", encoding="utf-8") as file_:
        json.dump(dict_, file_, ensure_ascii=False)

    return


# yml
def _load_yml(path: str) -> dict:
    with open(path, "r") as file_:
        return dict(yaml.safe_load(file_))


def _save_yml(dict_: dict, path: str) -> None:

    with open(path, "w", encoding="utf-8") as file_:
        return yaml.dump(dict_, file_, default_flow_style=False, allow_unicode=True)



"""
atpthings
=========

ATP Things python package

"""
from . import dict
from . import util
from .database.data_entity import DataEntity
from .database.data_timeseries import DataTimeseries

from . import _version

__version__ = _version.get_versions()["version"]
